"""Plotter module for creating plots of the molecule and it's orbitals."""

import logging
import tkinter as tk
from tkinter import ttk
from typing import Optional

import numpy as np
import pyvista as pv
from numpy.typing import NDArray
from pyvistaqt import BackgroundPlotter

from ._plotting_objects import Molecule
from .parser import _MolecularOrbital
from .tabulator import GridType, Tabulator, _cartesian_to_spherical, _spherical_to_cartesian

logger = logging.getLogger(__name__)


class Plotter:
    """
    Handles the 3D visualization of molecules and molecular orbitals.

    This class uses PyVista for 3D rendering and Tkinter for the user interface
    to control plotting parameters and select orbitals.

    Args
    ----
        source: str | list[str]
            The path to the molden file, or the lines from the file.

        only_molecule: bool, optional
            Only parse the atoms and skip molecular orbitals.
            Default is `False`.

        tabulator: Tabulator, optional
            If `None`, `Plotter` creates a `Tabulator` and tabulates the GTOs and MOs with a default grid.
            A `Tabulator` can be passed as to tabulate the GTOs in a predetermined grid.

            Note: `Tabulator` grid must be spherical or cartesian. Custom grids are not allowed.

        tk_rook: tk.Tk, optional
            If user is using the plotter inside a tk app, `tk_root` can be passed as to not create a new tk instance.

    Raises
    ------
        ValueError:
            If the provided tabulator is invalid
            (e.g., missing grid or GTO data when `only_molecule` is `False`, or has an UNKNOWN grid type).
    """

    # Default values
    CONTOUR = 0.1
    OPACITY = 1.0
    MOLECULE_OPACITY = 1.0

    NUM_RADIUS_POINTS = 100
    NUM_THETA_POINTS = 60
    NUM_PHI_POINTS = 120

    NUM_X_POINTS = 100
    NUM_Y_POINTS = 100
    NUM_Z_POINTS = 100

    def __init__(
        self,
        source: str | list[str],
        only_molecule: bool = False,
        tabulator: Optional[Tabulator] = None,
        tk_root: Optional[tk.Tk] = None,
    ) -> None:
        self.on_screen = True

        if tabulator:
            if not hasattr(tabulator, 'grid'):
                raise ValueError('Tabulator does not have grid attribute.')

            if not hasattr(tabulator, 'gto_data') and not only_molecule:
                raise ValueError('Tabulator does not have tabulated GTOs.')

            if tabulator.grid_type == GridType.UNKNOWN:
                raise ValueError('The plotter only supports spherical and cartesian grids.')

            self.tab = tabulator
        else:
            self.tab = Tabulator(source, only_molecule=only_molecule)

        self.molecule = Molecule(self.tab.atoms)
        self.molecule_opacity = self.MOLECULE_OPACITY

        if not only_molecule:
            self.tk_root = tk_root
            self.no_prev_root = self.tk_root is None
            if self.no_prev_root:
                self.tk_root = tk.Tk()
                self.tk_root.withdraw()  # Hides window

        self.pv_plotter = BackgroundPlotter(editor=False)
        self.pv_plotter.show_axes()
        self.molecule_actors = self.molecule.add_meshes(self.pv_plotter, self.molecule_opacity)

        # If we want to have the molecular orbitals, we need to initiate Tk before Qt
        # That is why we have this weird if statement separated this way
        if only_molecule:
            self.pv_plotter.app.exec_()  # pyright: ignore[reportAttributeAccessIssue]
            return

        assert self.tk_root is not None  # To help type hinters

        if not tabulator:
            # Default is a spherical grid
            self.tab.spherical_grid(
                np.linspace(0, 2 * self.molecule.max_radius, self.NUM_RADIUS_POINTS),
                np.linspace(0, np.pi, self.NUM_THETA_POINTS),
                np.linspace(0, 2 * np.pi, self.NUM_PHI_POINTS),
            )

        self.orb_mesh = self._create_mo_mesh()
        self.orb_actor: pv.Actor | None = None

        # Values for MO, not the molecule
        self.contour = self.CONTOUR
        self.opacity = self.OPACITY

        _OrbitalSelectionScreen(self, self.tk_root)

        self.tk_root.mainloop()

    def toggle_molecule(self) -> None:
        """Toggle the visibility of the molecule."""
        if self.molecule_actors:
            for actor in self.molecule_actors:
                actor.SetVisibility(not actor.GetVisibility())
            self.pv_plotter.update()

    def _create_mo_mesh(self) -> pv.StructuredGrid:
        """Create a mesh for the orbitals.

        Returns
        -------
            pv.StructuredGrid:
                The mesh object for MO visualization.

        """
        mesh = pv.StructuredGrid()
        mesh.points = pv.pyvista_ndarray(self.tab.grid)

        # Pyvista needs the dimensions backwards
        # in other words, (phi, theta, r) or (z, y, x)
        mesh.dimensions = self.tab.grid_dimensions[::-1]

        return mesh

    def update_mesh(
        self,
        i_points: NDArray[np.floating],
        j_points: NDArray[np.floating],
        k_points: NDArray[np.floating],
        grid_type: GridType,
    ) -> None:
        """Update the grid in the Tabulator and recreates the MO mesh.

        Args:
            i_points: NDArray[np.floating]:
                Points for the first dimension (radius or x).
            j_points: NDArray[np.floating]:
                Points for the second dimension (theta or y).
            k_points: NDArray[np.floating]:
                Points for the third dimension (phi or z).
            grid_type: GridType:
                The type of grid to create (SPHERICAL or CARTESIAN).

        Raises
        ------
            ValueError: If the grid_type is not SPHERICAL or CARTESIAN.
        """
        if grid_type == GridType.CARTESIAN:
            self.tab.cartesian_grid(i_points, j_points, k_points)
        elif grid_type == GridType.SPHERICAL:
            self.tab.spherical_grid(i_points, j_points, k_points)
        else:
            raise ValueError('The plotter only supports spherical and cartesian grids.')

        self.orb_mesh = self._create_mo_mesh()


class _OrbitalSelectionScreen(tk.Toplevel):
    SPHERICAL_GRID_SETTINGS_WINDOW_SIZE = '600x500'
    CARTESIAN_GRID_SETTINGS_WINDOW_SIZE = '800x500'

    def __init__(self, plotter: Plotter, tk_master: tk.Tk) -> None:
        super().__init__(tk_master)
        self.title('Orbitals')
        self.geometry('350x500')

        self.protocols()

        self.plotter = plotter
        self.current_orb_ind = -1  # Start with no orbital shown

        nav_frame = ttk.Frame(self)
        nav_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)

        self.prev_button = ttk.Button(nav_frame, text='<< Previous', command=self.prev_plot)
        self.prev_button.pack(side=tk.LEFT, padx=5, pady=10)

        self.next_button = ttk.Button(nav_frame, text='Next >>', command=self.next_plot)
        self.next_button.pack(side=tk.RIGHT, padx=5, pady=10)

        self.update_button_states()  # Update buttons for initial state

        self.settings_button = ttk.Button(self, text='Settings', command=self.settings_screen)
        self.settings_button.pack(expand=True, padx=5, pady=10)

        self.orb_tv = _OrbitalsTreeview(self)
        self.orb_tv.populate_tree(self.plotter.tab.mos)
        self.orb_tv.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)

    def on_close(self) -> None:
        self.plotter.on_screen = False
        self.plotter.pv_plotter.close()
        self.destroy()
        if self.plotter.tk_root and self.plotter.no_prev_root:
            self.plotter.tk_root.destroy()

    def protocols(self) -> None:
        self.protocol('WM_DELETE_WINDOW', self.on_close)
        self.bind('<Command-q>', lambda _event: self.on_close())
        self.bind('<Command-w>', lambda _event: self.on_close())
        self.bind('<Control-q>', lambda _event: self.on_close())
        self.bind('<Control-w>', lambda _event: self.on_close())

    def settings_screen(self) -> None:
        self.settings_window = tk.Toplevel(self)
        self.settings_window.title('Settings')

        settings_frame = ttk.Frame(self.settings_window)
        settings_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)

        # Contour level
        contour_label = ttk.Label(settings_frame, text='Molecular Orbital Contour:')
        contour_label.grid(row=0, column=0, padx=5, pady=5)
        self.contour_entry = ttk.Entry(settings_frame)
        self.contour_entry.insert(0, str(self.plotter.contour))
        self.contour_entry.grid(row=1, column=0, padx=5, pady=5)

        # Opacity
        opacity_label = ttk.Label(settings_frame)
        opacity_label.grid(row=2, column=0, padx=5, pady=5)
        self.opacity_scale = ttk.Scale(
            settings_frame,
            length=150,
            command=lambda val: opacity_label.config(text=f'Molecular Orbital Opacity: {float(val):.2f}'),
        )
        self.opacity_scale.set(self.plotter.opacity)
        self.opacity_scale.grid(row=3, column=0, padx=5, pady=5)

        # Molecule Opacity
        molecule_opacity_label = ttk.Label(settings_frame)
        molecule_opacity_label.grid(row=4, column=0, padx=5, pady=5)
        self.molecule_opacity_scale = ttk.Scale(
            settings_frame,
            length=150,
            command=lambda val: molecule_opacity_label.config(text=f'Molecule Opacity: {float(val):.2f}'),
        )
        self.molecule_opacity_scale.set(self.plotter.molecule_opacity)
        self.molecule_opacity_scale.grid(row=5, column=0, padx=5, pady=5)

        # Toggle molecule visibility
        toggle_mol_button = ttk.Button(settings_frame, text='Toggle Molecule', command=self.plotter.toggle_molecule)
        toggle_mol_button.grid(row=6, column=0, padx=5, pady=5)

        # Grid parameters
        ttk.Label(settings_frame, text='MO Grid parameters').grid(row=0, column=1, padx=5, pady=5, columnspan=4)

        self.grid_type_radio_var = tk.StringVar()
        self.grid_type_radio_var.set(self.plotter.tab.grid_type.value)

        ttk.Label(settings_frame, text='Spherical grid:').grid(row=1, column=1, padx=5, pady=5)
        sph_grid_type_button = ttk.Radiobutton(
            settings_frame,
            variable=self.grid_type_radio_var,
            value=GridType.SPHERICAL.value,
            command=self.place_grid_params_frame,
        )

        ttk.Label(settings_frame, text='Cartesian grid:').grid(row=1, column=3, padx=5, pady=5)
        cart_grid_type_button = ttk.Radiobutton(
            settings_frame,
            variable=self.grid_type_radio_var,
            value=GridType.CARTESIAN.value,
            command=self.place_grid_params_frame,
        )

        sph_grid_type_button.grid(row=1, column=2, padx=5, pady=5)
        cart_grid_type_button.grid(row=1, column=4, padx=5, pady=5)

        self.sph_grid_params_frame = self.sph_grid_params_frame_widgets(settings_frame)
        self.cart_grid_params_frame = self.cart_grid_params_frame_widgets(settings_frame)

        self.place_grid_params_frame()

        # Reset button
        reset_button = ttk.Button(settings_frame, text='Reset', command=self.reset_settings)
        reset_button.grid(row=8, column=0, padx=5, pady=5, columnspan=5)

        # Save settings button
        save_button = ttk.Button(settings_frame, text='Apply', command=self.apply_settings)
        save_button.grid(row=9, column=0, padx=5, pady=5, columnspan=5)

    def place_grid_params_frame(self) -> None:
        if self.grid_type_radio_var.get() == GridType.SPHERICAL.value:
            self.settings_window.geometry(self.SPHERICAL_GRID_SETTINGS_WINDOW_SIZE)
            self.cart_grid_params_frame.grid_forget()
            self.settings_window.geometry()
            self.sph_grid_params_frame.grid(row=2, column=1, padx=5, pady=5, rowspan=6, columnspan=4)
            self.sph_grid_params_frame_setup()
        else:
            self.settings_window.geometry(self.CARTESIAN_GRID_SETTINGS_WINDOW_SIZE)
            self.sph_grid_params_frame.grid_forget()
            self.cart_grid_params_frame.grid(row=2, column=1, padx=5, pady=5, rowspan=6, columnspan=4)
            self.cart_grid_params_frame_setup()

    def sph_grid_params_frame_widgets(self, master: ttk.Frame) -> ttk.Frame:
        grid_params_frame = ttk.Frame(master)

        # Radius
        ttk.Label(grid_params_frame, text='Radius:').grid(row=0, column=0, padx=5, pady=5)
        self.radius_entry = ttk.Entry(grid_params_frame)
        self.radius_entry.grid(row=0, column=1, padx=5, pady=5)

        # Radius points
        radius_points_label = ttk.Label(grid_params_frame, text='Number of Radius Points:')
        radius_points_label.grid(row=1, column=0, padx=5, pady=5)
        self.radius_points_entry = ttk.Entry(grid_params_frame)
        self.radius_points_entry.grid(row=1, column=1, padx=5, pady=5)

        # Theta points
        theta_points_label = ttk.Label(grid_params_frame, text='Number of Theta Points:')
        theta_points_label.grid(row=2, column=0, padx=5, pady=5)
        self.theta_points_entry = ttk.Entry(grid_params_frame)
        self.theta_points_entry.grid(row=2, column=1, padx=5, pady=5)

        # Phi points
        phi_points_label = ttk.Label(grid_params_frame, text='Number of Phi Points:')
        phi_points_label.grid(row=3, column=0, padx=5, pady=5)
        self.phi_points_entry = ttk.Entry(grid_params_frame)
        self.phi_points_entry.grid(row=3, column=1, padx=5, pady=5)

        return grid_params_frame

    def cart_grid_params_frame_widgets(self, master: ttk.Frame) -> ttk.Frame:
        grid_params_frame = ttk.Frame(master)

        # X
        ttk.Label(grid_params_frame, text='Min x:').grid(row=0, column=0, padx=5, pady=5)
        ttk.Label(grid_params_frame, text='Max x:').grid(row=0, column=1, padx=5, pady=5)
        ttk.Label(grid_params_frame, text='Num x points:').grid(row=0, column=2, padx=5, pady=5)

        self.x_min_entry = ttk.Entry(grid_params_frame)
        self.x_max_entry = ttk.Entry(grid_params_frame)
        self.x_num_points_entry = ttk.Entry(grid_params_frame)

        self.x_min_entry.grid(row=1, column=0, padx=5, pady=5)
        self.x_max_entry.grid(row=1, column=1, padx=5, pady=5)
        self.x_num_points_entry.grid(row=1, column=2, padx=5, pady=5)

        # Y
        ttk.Label(grid_params_frame, text='Min y:').grid(row=2, column=0, padx=5, pady=5)
        ttk.Label(grid_params_frame, text='Max y:').grid(row=2, column=1, padx=5, pady=5)
        ttk.Label(grid_params_frame, text='Num y points:').grid(row=2, column=2, padx=5, pady=5)

        self.y_min_entry = ttk.Entry(grid_params_frame)
        self.y_max_entry = ttk.Entry(grid_params_frame)
        self.y_num_points_entry = ttk.Entry(grid_params_frame)

        self.y_min_entry.grid(row=3, column=0, padx=5, pady=5)
        self.y_max_entry.grid(row=3, column=1, padx=5, pady=5)
        self.y_num_points_entry.grid(row=3, column=2, padx=5, pady=5)

        # Z
        ttk.Label(grid_params_frame, text='Min z:').grid(row=4, column=0, padx=5, pady=5)
        ttk.Label(grid_params_frame, text='Max z:').grid(row=4, column=1, padx=5, pady=5)
        ttk.Label(grid_params_frame, text='Num z points:').grid(row=4, column=2, padx=5, pady=5)

        self.z_min_entry = ttk.Entry(grid_params_frame)
        self.z_max_entry = ttk.Entry(grid_params_frame)
        self.z_num_points_entry = ttk.Entry(grid_params_frame)

        self.z_min_entry.grid(row=5, column=0, padx=5, pady=5)
        self.z_max_entry.grid(row=5, column=1, padx=5, pady=5)
        self.z_num_points_entry.grid(row=5, column=2, padx=5, pady=5)

        return grid_params_frame

    def sph_grid_params_frame_setup(self) -> None:
        self.radius_entry.delete(0, tk.END)
        self.radius_points_entry.delete(0, tk.END)
        self.theta_points_entry.delete(0, tk.END)
        self.phi_points_entry.delete(0, tk.END)

        # Previous grid was cartesian, so use default values
        if self.plotter.tab.grid_type == GridType.CARTESIAN:
            self.radius_entry.insert(0, str(self.plotter.molecule.max_radius * 2))
            self.radius_points_entry.insert(0, str(self.plotter.NUM_RADIUS_POINTS))
            self.theta_points_entry.insert(0, str(self.plotter.NUM_THETA_POINTS))
            self.phi_points_entry.insert(0, str(self.plotter.NUM_PHI_POINTS))
            return

        num_r, num_theta, num_phi = self.plotter.tab.grid_dimensions

        # The last point of the grid for sure has the largest r
        r, _, _ = _cartesian_to_spherical(*self.plotter.tab.grid[-1, :])

        self.radius_entry.insert(0, str(r))
        self.radius_points_entry.insert(0, str(num_r))
        self.theta_points_entry.insert(0, str(num_theta))
        self.phi_points_entry.insert(0, str(num_phi))

    def cart_grid_params_frame_setup(self) -> None:
        self.x_min_entry.delete(0, tk.END)
        self.x_max_entry.delete(0, tk.END)
        self.x_num_points_entry.delete(0, tk.END)

        self.y_min_entry.delete(0, tk.END)
        self.y_max_entry.delete(0, tk.END)
        self.y_num_points_entry.delete(0, tk.END)

        self.z_min_entry.delete(0, tk.END)
        self.z_max_entry.delete(0, tk.END)
        self.z_num_points_entry.delete(0, tk.END)

        # Previous grid was sphesical, so use adapted default values
        if self.plotter.tab.grid_type == GridType.SPHERICAL:
            r = 2 * self.plotter.molecule.max_radius

            self.x_min_entry.insert(0, str(-r))
            self.y_min_entry.insert(0, str(-r))
            self.z_min_entry.insert(0, str(-r))

            self.x_max_entry.insert(0, str(r))
            self.y_max_entry.insert(0, str(r))
            self.z_max_entry.insert(0, str(r))

            self.x_num_points_entry.insert(0, str(self.plotter.NUM_X_POINTS))
            self.y_num_points_entry.insert(0, str(self.plotter.NUM_Y_POINTS))
            self.z_num_points_entry.insert(0, str(self.plotter.NUM_Z_POINTS))
            return

        x_num, y_num, z_num = self.plotter.tab.grid_dimensions
        x_min, y_min, z_min = self.plotter.tab.grid[0, :]
        x_max, y_max, z_max = self.plotter.tab.grid[-1, :]

        self.x_min_entry.insert(0, str(x_min))
        self.x_max_entry.insert(0, str(x_max))
        self.x_num_points_entry.insert(0, str(x_num))

        self.y_min_entry.insert(0, str(y_min))
        self.y_max_entry.insert(0, str(y_max))
        self.y_num_points_entry.insert(0, str(y_num))

        self.z_min_entry.insert(0, str(z_min))
        self.z_max_entry.insert(0, str(z_max))
        self.z_num_points_entry.insert(0, str(z_num))

    def reset_settings(self) -> None:
        """Reset settings to default values."""
        self.contour_entry.delete(0, tk.END)
        self.contour_entry.insert(0, str(self.plotter.CONTOUR))

        self.opacity_scale.set(Plotter.OPACITY)

        self.molecule_opacity_scale.set(self.plotter.MOLECULE_OPACITY)

        self.grid_type_radio_var.set(GridType.SPHERICAL.value)

        self.radius_entry.delete(0, tk.END)
        self.radius_entry.insert(0, str(self.plotter.molecule.max_radius * 2))

        self.radius_points_entry.delete(0, tk.END)
        self.radius_points_entry.insert(0, str(self.plotter.NUM_RADIUS_POINTS))

        self.theta_points_entry.delete(0, tk.END)
        self.theta_points_entry.insert(0, str(self.plotter.NUM_THETA_POINTS))

        self.phi_points_entry.delete(0, tk.END)
        self.phi_points_entry.insert(0, str(self.plotter.NUM_PHI_POINTS))

    def apply_settings(self) -> None:
        self.plotter.molecule_opacity = round(self.molecule_opacity_scale.get(), 2)
        for actor in self.plotter.molecule_actors:
            actor.GetProperty().SetOpacity(self.plotter.molecule_opacity)

        if self.grid_type_radio_var.get() == GridType.SPHERICAL.value:
            radius = float(self.radius_entry.get())
            if radius <= 0:
                raise ValueError('Invalid radius value: Must be greater than zero.')

            num_r_points = int(self.radius_points_entry.get())
            num_theta_points = int(self.theta_points_entry.get())
            num_phi_points = int(self.phi_points_entry.get())

            if num_r_points <= 0 or num_theta_points <= 0 or num_phi_points <= 0:
                raise ValueError('Invalid number of points: Must be greater than zero.')

            r = np.linspace(0, radius, num_r_points)
            theta = np.linspace(0, np.pi, num_theta_points)
            phi = np.linspace(0, 2 * np.pi, num_phi_points)

            rr, tt, pp = np.meshgrid(r, theta, phi, indexing='ij')
            xx, yy, zz = _spherical_to_cartesian(rr, tt, pp)

            # Update the mesh with new points if needed
            new_grid = np.column_stack((xx.ravel(), yy.ravel(), zz.ravel()))
            if not np.array_equal(new_grid, self.plotter.tab.grid):
                self.plotter.update_mesh(r, theta, phi, GridType.SPHERICAL)

        else:
            x_min = float(self.x_min_entry.get())
            x_max = float(self.x_max_entry.get())
            x_num = int(self.x_num_points_entry.get())

            y_min = float(self.y_min_entry.get())
            y_max = float(self.y_max_entry.get())
            y_num = int(self.y_num_points_entry.get())

            z_min = float(self.z_min_entry.get())
            z_max = float(self.z_max_entry.get())
            z_num = int(self.z_num_points_entry.get())

            if x_num <= 0 or y_num <= 0 or z_num <= 0:
                raise ValueError('Invalid number of points: Must be greater than zero.')

            x = np.linspace(x_min, x_max, x_num)
            y = np.linspace(y_min, y_max, y_num)
            z = np.linspace(z_min, z_max, z_num)

            xx, yy, zz = np.meshgrid(x, y, z, indexing='ij')

            # Update the mesh with new points if needed
            new_grid = np.column_stack((xx.ravel(), yy.ravel(), zz.ravel()))
            if not np.array_equal(new_grid, self.plotter.tab.grid):
                self.plotter.update_mesh(x, y, z, GridType.CARTESIAN)

        self.plotter.contour = float(self.contour_entry.get().strip())
        self.plot_orbital(self.current_orb_ind)

        self.plotter.opacity = round(self.opacity_scale.get(), 2)
        if self.plotter.orb_actor:
            self.plotter.orb_actor.GetProperty().SetOpacity(self.plotter.opacity)

    def next_plot(self) -> None:
        """Go to the next plot."""
        self.current_orb_ind += 1
        self.update_button_states()
        self.orb_tv.highlight_orbital(self.current_orb_ind)
        self.plot_orbital(self.current_orb_ind)

    def prev_plot(self) -> None:
        """Go to the previous plot."""
        self.current_orb_ind -= 1
        self.orb_tv.highlight_orbital(self.current_orb_ind)
        self.update_button_states()
        self.plot_orbital(self.current_orb_ind)

    def update_button_states(self) -> None:
        """Update the enabled/disabled state of nav buttons."""
        can_go_prev = self.current_orb_ind > 0
        can_go_next = self.current_orb_ind < len(self.plotter.tab.mos) - 1
        self.prev_button.config(state=tk.NORMAL if can_go_prev else tk.DISABLED)
        self.next_button.config(state=tk.NORMAL if can_go_next else tk.DISABLED)

    def plot_orbital(self, orb_ind: int) -> None:
        if self.plotter.orb_actor:
            self.plotter.pv_plotter.remove_actor(self.plotter.orb_actor)

        if orb_ind != -1:
            self.plotter.orb_mesh['orbital'] = self.plotter.tab.tabulate_mos(orb_ind)

            contour_mesh = self.plotter.orb_mesh.contour([-self.plotter.contour, self.plotter.contour])

            self.plotter.orb_actor = self.plotter.pv_plotter.add_mesh(
                contour_mesh,
                clim=[-self.plotter.contour, self.plotter.contour],
                opacity=self.plotter.opacity,
                show_scalar_bar=False,
                cmap='bwr',
            )


class _OrbitalsTreeview(ttk.Treeview):
    def __init__(self, selection_screen: _OrbitalSelectionScreen) -> None:
        columns = ['Index', 'Symmetry', 'Occupation', 'Energy [au]']
        widths = [20, 50, 50, 120]

        super().__init__(selection_screen, columns=columns, show='headings', height=20)

        for col, w in zip(columns, widths):
            self.heading(col, text=col)
            self.column(col, width=w)

        self.selection_screen = selection_screen

        self.current_orb_ind = -1  # Start with no orbital shown

        # Configure tag
        self.tag_configure('highlight', background='lightblue')

        self.bind('<<TreeviewSelect>>', self.on_select)

    def highlight_orbital(self, orb_ind: int) -> None:
        """Highlight the selected orbital."""
        if self.current_orb_ind != -1:
            self.item(self.current_orb_ind, tags=('!hightlight',))

        self.current_orb_ind = orb_ind
        self.item(orb_ind, tags=('highlight',))
        self.see(orb_ind)  # Scroll to the selected item

    def erase(self) -> None:
        for item in self.get_children():
            self.delete(item)

    def populate_tree(self, mos: list[_MolecularOrbital]) -> None:
        self.erase()

        # Counts the number of MOs with a given symmetry
        mo_syms = list({mo.sym for mo in mos})
        mo_sym_count: dict[str, int] = dict.fromkeys(mo_syms, 0)
        for ind, mo in enumerate(mos):
            mo_sym_count[mo.sym] += 1
            self.insert('', 'end', iid=ind, values=(ind + 1, f'{mo.sym}.{mo_sym_count[mo.sym]}', mo.occ, mo.energy))

    def on_select(self, _event: tk.Event) -> None:
        """Handle selection of an orbital."""
        selected_item = self.selection()
        self.selection_remove(selected_item)
        if selected_item:
            orb_ind = int(selected_item[0])
            self.highlight_orbital(orb_ind)
            self.selection_screen.current_orb_ind = orb_ind
            self.selection_screen.plot_orbital(orb_ind)
            self.selection_screen.update_button_states()



# ruff: noqa
import numpy as np
import pytest
from pathlib import Path

from moldenViz.tabulator import Tabulator, _spherical_to_cartesian, _cartesian_to_spherical, array_like_type

MOLDEN_PATH = Path(__file__).with_name('sample_molden.inp')


def test_spherical_cartesian_roundtrip() -> None:
    rng = np.random.default_rng(seed=42)
    r_vals = rng.uniform(0.1, 10.0, size=100)
    theta_vals = rng.uniform(0.0, np.pi, size=100)
    phi_vals = rng.uniform(-np.pi, np.pi, size=100)

    x, y, z = _spherical_to_cartesian(r_vals, theta_vals, phi_vals)
    r2, theta2, phi2 = _cartesian_to_spherical(x, y, z)

    np.testing.assert_allclose(r_vals, r2, rtol=1e-12, atol=1e-12)
    np.testing.assert_allclose(theta_vals, theta2, rtol=1e-12, atol=1e-12)
    assert np.allclose(phi_vals, phi2)


def test_tabulate_gtos_requires_grid() -> None:
    tab = Tabulator(str(MOLDEN_PATH))
    with pytest.raises(ValueError):
        tab.tabulate_gtos()


def test_cartesian_grid_shape() -> None:
    tab = Tabulator(str(MOLDEN_PATH))
    x, y, z = np.linspace(-1, 1, 3), np.linspace(-1, 1, 4), np.linspace(-1, 1, 2)
    tab.cartesian_grid(x, y, z, tabulate_gtos=False)
    assert tab.grid is not None
    assert tab.grid.shape == (len(x) * len(y) * len(z), 3)


def test_spherical_grid_shape() -> None:
    tab = Tabulator(str(MOLDEN_PATH))
    r, theta, phi = np.r_[1.0, 2.0], np.r_[0.0, np.pi / 2, np.pi], np.r_[-np.pi, 0.0, np.pi / 2, np.pi]
    tab.spherical_grid(r, theta, phi, tabulate_gtos=False)
    assert tab.grid is not None
    assert tab.grid.shape == (len(r) * len(theta) * len(phi), 3)


@pytest.mark.parametrize('lmax, num_points', [(0, 10), (3, 25), (5, 50)])
def test_tabulate_xlms_shape(lmax: int, num_points: int) -> None:
    theta = np.linspace(0.0, np.pi, num_points, dtype=float)
    phi = np.linspace(-np.pi, np.pi, num_points, dtype=float)

    xlms = Tabulator._tabulate_xlms(theta, phi, lmax)

    assert xlms.shape == (lmax + 1, 2 * lmax + 1, num_points)


@pytest.mark.parametrize('mo_inds', [None, 0, [0], [0, 1, 2], [0, 1, 2, 3, 4], range(1, 10)])
def test_tabulate_mos(mo_inds: int | array_like_type | None) -> None:
    tab = Tabulator(str(MOLDEN_PATH))
    tab.cartesian_grid(np.linspace(-1, 1, 5), np.linspace(-1, 1, 5), np.linspace(-1, 1, 5))
    mo_data = tab.tabulate_mos(mo_inds)

    assert mo_data is not None

    if mo_inds is None:
        assert mo_data.shape == (125, 177)
    elif isinstance(mo_inds, int):
        assert mo_data.shape == (125,)
    else:
        assert mo_data.shape == (125, len(mo_inds))


@pytest.mark.parametrize('mo_inds', [-1, range(0), range(-1, 1), [0, -1], [1, 2, 3, -1], [0, 178]])
def test_invalid_mo_inds(mo_inds: int | array_like_type | None) -> None:
    tab = Tabulator(str(MOLDEN_PATH))
    tab.cartesian_grid(np.linspace(-1, 1, 5), np.linspace(-1, 1, 5), np.linspace(-1, 1, 5))

    with pytest.raises(ValueError):
        tab.tabulate_mos(mo_inds)



"""Tabulator module for creating grids and tabulating Gaussian-type orbitals (GTOs) from Molden files."""

import logging
from enum import Enum
from math import factorial
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from .parser import Parser

logger = logging.getLogger(__name__)

array_like_type = NDArray[np.integer] | list[int] | tuple[int, ...] | range


def _spherical_to_cartesian(
    r: NDArray[np.floating],
    theta: NDArray[np.floating],
    phi: NDArray[np.floating],
) -> tuple[NDArray[np.floating], NDArray[np.floating], NDArray[np.floating]]:
    """Convert spherical coordinates to Cartesian coordinates.

    Args
    ----
        r (NDArray[np.floating]): Radial distances.
        theta (NDArray[np.floating]): Polar angles (in radians).
        phi (NDArray[np.floating]): Azimuthal angles (in radians).

    Returns
    -------
        tuple: Arrays of x, y, z Cartesian coordinates.

    """
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    return x, y, z


def _cartesian_to_spherical(
    x: NDArray[np.floating],
    y: NDArray[np.floating],
    z: NDArray[np.floating],
) -> tuple[NDArray[np.floating], NDArray[np.floating], NDArray[np.floating]]:
    """Convert Cartesian coordinates to spherical coordinates.

    Args
    ----
        x (NDArray[np.floating]): X coordinates.
        y (NDArray[np.floating]): Y coordinates.
        z (NDArray[np.floating]): Z coordinates.

    Returns
    -------
        tuple: Arrays of r (radius), theta (polar angle), phi (azimuthal angle).

    """
    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arccos(z / r)
    phi = np.arctan2(y, x)

    return r, theta, phi


def _binomial(r: float, k: int) -> float:
    """
    Calculate the generalized binomial coefficient (r over k).

    This function supports real or complex 'r' and non-negative integer 'k'.
    The formula is: C(r, k) = r * (r-1) * ... * (r-k+1) / k!

    Args:
        r: A real or complex number.
        k: A non-negative integer.

    Returns
    -------
        The value of the generalized binomial coefficient as a float or complex number.

    Raises
    ------
        ValueError: If k is a negative integer.
    """
    if not isinstance(k, int) or k < 0:
        raise ValueError('k must be a non-negative integer.')

    if k == 0:
        return 1
    if k == 1:
        return r

    numerator = 1
    for i in range(k):
        numerator *= r - i

    denominator = factorial(k)

    return numerator / denominator


class GridType(Enum):
    """Grid types allowed."""

    SPHERICAL = 'spherical'
    CARTESIAN = 'cartesian'
    UNKNOWN = 'unknown'


class Tabulator(Parser):
    """Extends Parses, create grids and tabulates Gaussian-type orbitals (GTOs) from Molden files.

    Args
    ----
        source: str | list[str]
            The path to the molden file, or the lines from the file.

        only_molecule: bool, optional
            Only parse the atoms and skip molecular orbitals.
            Default is `False`.
    """

    def __init__(
        self,
        source: str | list[str],
        only_molecule: bool = False,
    ) -> None:
        """Initialize the Tabulator with a Molden file or its content."""
        super().__init__(source, only_molecule)

        self.only_molecule = only_molecule
        self.grid_type = GridType.UNKNOWN

        self.grid: NDArray[np.floating]
        self.grid_dimensions: tuple[int, int, int]
        self.gtos_data: NDArray[np.floating]
        self.mos_data: NDArray[np.floating]

    def cartesian_grid(
        self,
        x: NDArray[np.floating],
        y: NDArray[np.floating],
        z: NDArray[np.floating],
        tabulate_gtos: bool = True,
    ) -> None:
        r"""Create cartesian grid from x, y, z arrays and tabulate GTOs.

        Args
        ----
            x: NDArray[np.floating]
                Array of x coordinates.
            y: NDArray[np.floating]
                Array of y coordinates.
            z: NDArray[np.floating]
                Array of z coordinates.
            tabulate_gtos: bool, optional
                Whether to tabulate Gaussian-type orbitals (GTOs) after creating the grid.
                Defaults to True.

        """
        if self.only_molecule:
            raise RuntimeError('Grid creation is not allowed when `only_molecule` is set to `True`.')

        xx, yy, zz = np.meshgrid(x, y, z, indexing='ij')
        self.grid = np.column_stack((xx.ravel(), yy.ravel(), zz.ravel()))
        self.grid_type = GridType.CARTESIAN
        self.grid_dimensions = (len(x), len(y), len(z))

        if tabulate_gtos:
            self.gtos_data = self.tabulate_gtos()

    def spherical_grid(
        self,
        r: NDArray[np.floating],
        theta: NDArray[np.floating],
        phi: NDArray[np.floating],
        tabulate_gtos: bool = True,
    ) -> None:
        r"""Create spherical grid from r, theta, phi arrays and tabulate GTOs.

        Args
        ----
            r: NDArray[np.floating]
                Array of radial coordinates.
            theta: NDArray[np.floating]
                Array of polar angles (in radians).
            phi: NDArray[np.floating]
                Array of azimuthal angles (in radians).
            tabulate_gtos: bool, optional
                Whether to tabulate Gaussian-type orbitals (GTOs) after creating the grid.
                Defaults to True.

        Note
        ----
            Grid points are converted to Cartesian coordinates.

        """
        if self.only_molecule:
            raise RuntimeError('Grid creation is not allowed when `only_molecule` is set to `True`.')

        rr, tt, pp = np.meshgrid(r, theta, phi, indexing='ij')
        xx, yy, zz = _spherical_to_cartesian(rr, tt, pp)
        self.grid = np.column_stack((xx.ravel(), yy.ravel(), zz.ravel()))
        self.grid_type = GridType.SPHERICAL
        self.grid_dimensions = (len(r), len(theta), len(phi))

        if tabulate_gtos:
            self.gtos_data = self.tabulate_gtos()

    def tabulate_gtos(self) -> NDArray[np.floating]:
        """Tabulate Gaussian-type orbitals (GTOs) on the current grid.

        Returns
        -------
            NDArray[np.floating]:
                Array containing the tabulated GTOs data.

        Raises
        ------
            ValueError:
                If the grid is not defined before tabulating GTOs.

        """
        if self.only_molecule:
            raise RuntimeError('Grid creation is not allowed when `only_molecule` is set to `True`.')

        if not hasattr(self, 'grid'):
            raise ValueError('Grid is not defined. Please create a grid before tabulating GTOs.')

        # Having a predefined array makes it faster to fill the data
        gto_data = np.empty((self.grid.shape[0], len(self.mos[0].coeffs)))
        ind = 0
        for atom in self.atoms:
            centered_grid = self.grid - atom.position
            max_l = atom.shells[-1].l

            r, theta, phi = _cartesian_to_spherical(*centered_grid.T)
            xlms = self._tabulate_xlms(theta, phi, max_l)

            for shell in atom.shells:
                l = shell.l
                m_inds = np.arange(-l, l + 1)
                gto_inds = ind + l + m_inds

                radial = shell.norm * r**l * sum(gto.norm * gto.coeff * np.exp(-gto.exp * r**2) for gto in shell.gtos)

                gto_data[:, gto_inds] = radial[:, None] * xlms[l, m_inds, ...].T

                ind += 2 * l + 1

        logger.debug('GTO data shape: %s', gto_data.shape)
        return gto_data

    def tabulate_mos(self, mo_inds: Optional[int | array_like_type] = None) -> NDArray[np.floating]:
        """Tabulate molecular orbitals (MOs) on the current grid.

        Args
        ----
            mo_inds : int, ArrayLike, None
                Indices of the MOs to tabulate. If None, all MOs are tabulated.

        Returns
        -------
            NDArray[np.floating]:
                Array containing the tabulated MOs data.

                If an integer is provided, it will tabulate only that MO.
                If an array-like is provided, it will tabulate the MOs at those indices.

        Raises
        ------
            ValueError:
                If the grid is not defined before tabulating MOs.
            ValueError:
                If GTOs are not tabulated before tabulating MOs.
            ValueError:
                If provided mo_inds invalid.

        """
        if not hasattr(self, 'grid'):
            raise ValueError('Grid is not defined. Please create a grid before tabulating MOs.')

        if not hasattr(self, 'gtos_data'):
            raise ValueError('GTOs are not tabulated. Please tabulate GTOs before tabulating MOs.')

        if mo_inds is None:
            mo_inds = list(range(len(self.mos)))

        if isinstance(mo_inds, range):
            mo_inds = list(mo_inds)

        if not isinstance(mo_inds, int) and not mo_inds:
            raise ValueError('Provided mo_inds is empty. Please provide valid indices.')

        if isinstance(mo_inds, int):
            if mo_inds < 0 or mo_inds >= len(self.mos):
                raise ValueError('Provided mo_index is invalid. Please provide valid index.')
        elif any(mo_ind < 0 or mo_ind >= len(self.mos) for mo_ind in mo_inds):
            raise ValueError('Provided mo_inds contains invalid indices. Please provide valid indices.')

        if isinstance(mo_inds, int):
            mo_data = np.sum(self.gtos_data * self.mos[mo_inds].coeffs[None, :], axis=1)
        else:
            mo_coeffs = np.stack([self.mos[i].coeffs for i in mo_inds])

            mo_data = np.sum(self.gtos_data[:, None, :] * mo_coeffs[None, ...], axis=2)
            logger.debug('MO data shape: %s', mo_data.shape)

        return mo_data

    @staticmethod
    def _tabulate_xlms(theta: NDArray[np.floating], phi: NDArray[np.floating], lmax: int) -> NDArray[np.floating]:
        r"""Tabulate the real spherical harmonics for given theta and phi values.

        We define the real spherical harmonics, Xlms
        (see eq.6, M.A. Blanco et al./Journal of Molecular Structure (Theochem) 419 (1997) 19-27), as:

        Xlms = sqrt(2)*Pl|m|s(\theta)*sin(|m|\phi), m<0
        Xlms = sqrt(2)*Plms(\theta)*cos(m\phi), m>0
        Xlms =         Plms             , m=0

        Note: Above, the Plms are normalized, i.e, \Theta_{lm}(\theta) in eq 1 of the paper.

        Args
        ----
            theta: NDArray[np.floating]
                Array of theta values.
            phi: NDArray[np.floating]
                Array of phi values.
            lmax: int
                Maximum angular momentum quantum number.

        Returns
        -------
            NDArray[np.floating]:
                Tabulated real spherical harmonics.

        """
        plms = Tabulator._tabulate_plms(np.cos(theta), lmax)

        xlms = np.empty((lmax + 1, 2 * lmax + 1, theta.size), dtype=float)
        xlms[:, 0, :] = plms[:, 0, :]

        for m in range(1, lmax + 1):
            xlms[:, -m, :] = np.sqrt(2) * plms[:, m, :] * np.sin(m * phi)
            xlms[:, m, :] = np.sqrt(2) * plms[:, m, :] * np.cos(m * phi)

        return xlms

    @staticmethod
    def _tabulate_plms(x: NDArray[np.floating], lmax: int) -> NDArray[np.floating]:
        """Tabulate normalized associated Legendre polynomials (without Condon-Shortley phase).

        Returns an array of shape (`lmax+1`, `lmax+1`, `x.size`), where:
            The first index is 0 <= l <= lmax
            The second index is 0 <= m <= lmax
            The third index goes over the x points

        Using closed form outlined here:
        https://en.m.wikipedia.org/wiki/Associated_Legendre_polynomials#Closed_Form

        Args
        ----
            x: NDArray[np.floating]
                Array of x values.
            lmax: int
                Maximum angular momentum quantum number.

        Returns
        -------
            NDArray[np.floating]:
                Tabulated associated Legendre polynomials.


        """
        plms = np.zeros((lmax + 1, lmax + 1, x.size), dtype=float)

        for l in range(lmax + 1):
            for m in range(l + 1):
                plms[l, m, :] = (
                    np.sqrt((2 * l + 1) * factorial(l - m) / factorial(l + m) / 4 / np.pi)
                    * 2**l
                    * (1 - x**2) ** (m / 2)
                    * sum(
                        factorial(k) * x ** (k - m) * _binomial(l, k) * _binomial((l + k - 1) / 2, l) / factorial(k - m)
                        for k in range(m, l + 1)
                    )
                )

        return plms



# SPDX-FileCopyrightText: 2025-present Faria22 <fariafelipe22@hotmail.com>
#
# SPDX-License-Identifier: MIT
from .parser import Parser
from .plotter import Plotter
from .tabulator import Tabulator

__all__ = ['Parser', 'Plotter', 'Tabulator']



# ruff: noqa
from pathlib import Path

import numpy as np
import pytest

from moldenViz.parser import Parser, _GTO, _Shell

# ----------------------------------------------------------------------
# utilities
# ----------------------------------------------------------------------
MOLDEN_PATH = Path(__file__).with_name('sample_molden.inp')


@pytest.fixture(scope='session')
def parser_obj() -> Parser:
    """Parser built once per test session from the reference Molden file."""
    return Parser(str(MOLDEN_PATH))


# ----------------------------------------------------------------------
# basic structural sanity
# ----------------------------------------------------------------------
def test_section_indices_order(parser_obj: Parser) -> None:
    assert parser_obj.atom_ind < parser_obj.gto_ind < parser_obj.mo_ind


def test_gaussian_normalization_positive() -> None:
    gto = _GTO(0.8, 0.5)
    gto.normalize(l=2)
    shell = _Shell(2, [gto])
    shell.normalize()
    assert gto.norm > 0.0 and shell.norm > 0.0


def test_atomic_orbital_permutation(parser_obj: Parser) -> None:
    order = parser_obj._gto_order()
    assert sorted(order) == list(range(len(order)))


def test_atom_labels(parser_obj: Parser) -> None:
    labels = [atm.label for atm in parser_obj.atoms]
    assert labels == ['Br', 'C_a', 'C_b', 'C_c', 'C_d', 'H']


def test_basis_and_mo_dimensions(parser_obj: Parser) -> None:
    n_basis = sum(2 * shell.l + 1 for shell in parser_obj.shells)

    assert len(parser_obj.mos) == 177

    # every MO coefficient vector must have that length
    for mo in parser_obj.mos:
        assert len(mo.coeffs) == n_basis


def test_mo_energies_are_sorted(parser_obj: Parser) -> None:
    energies = np.asarray([mo.energy for mo in parser_obj.mos])
    assert np.all(np.diff(energies) >= 0.0)


# ----------------------------------------------------------------------
# reproducibility checks
# ----------------------------------------------------------------------
def test_file_vs_lines_consistency(tmp_path) -> None:
    """Parsing via filename or via pre-read lines must give identical results."""
    lines = MOLDEN_PATH.read_text().splitlines(True)

    p_from_lines = Parser(lines)

    tmp_file = tmp_path / 'copy.molden'
    tmp_file.write_text(''.join(lines))
    p_from_file = Parser(str(tmp_file))

    # Quick invariants - if these match, deeper structures are identical
    assert [a.atomic_number for a in p_from_lines.atoms] == [a.atomic_number for a in p_from_file.atoms]
    assert [mo.energy for mo in p_from_lines.mos] == [mo.energy for mo in p_from_file.mos]



"""CLI entrance point."""

import argparse

from .examples.get_example_files import all_examples
from .plotter import Plotter


def main() -> None:
    parser = argparse.ArgumentParser(prog='moldenViz')
    source = parser.add_mutually_exclusive_group(required=True)

    source.add_argument('file', nargs='?', default=None, help='Optional molden file path', type=str)
    parser.add_argument('-m', '--only_molecule', action='store_true', help='Only plots the molecule')
    source.add_argument(
        '-e',
        '--example',
        nargs='?',
        type=str,
        metavar='molecule',
        choices=all_examples.keys(),
        help='Load example %(metavar)s. Options are: %(choices)s',
    )

    args = parser.parse_args()

    Plotter(
        args.file or all_examples[args.example],
        only_molecule=args.only_molecule,
    )


if __name__ == '__main__':
    main()



"""Get example files from folder and make them available to the package."""

from pathlib import Path


def _read_file(path: Path) -> list[str]:
    with path.open('r') as f:
        return f.readlines()


molden_files_folder = Path(__file__).parent / 'molden_files'

co = _read_file(molden_files_folder / 'co.inp')
o2 = _read_file(molden_files_folder / 'o2.inp')
co2 = _read_file(molden_files_folder / 'co2.inp')
h2o = _read_file(molden_files_folder / 'h2o.inp')
benzene = _read_file(molden_files_folder / 'benzene.inp')
prismane = _read_file(molden_files_folder / 'prismane.inp')
pyridine = _read_file(molden_files_folder / 'pyridine.inp')
furan = _read_file(molden_files_folder / 'furan.inp')
acrolein = _read_file(molden_files_folder / 'acrolein.inp')

all_examples = {
    'co': co,
    'o2': o2,
    'co2': co2,
    'h2o': h2o,
    'benzene': benzene,
    'prismane': prismane,
    'pyridine': pyridine,
    'furan': furan,
    'acrolein': acrolein,
}



import logging
from dataclasses import dataclass
from typing import cast

import numpy as np
import pyvista as pv
from numpy.typing import NDArray
from scipy.spatial.distance import pdist, squareform

from .parser import _Atom

logger = logging.getLogger(__name__)


@dataclass
class AtomType:
    name: str
    color: str
    radius: float
    max_bonds: int


# Atom type based on charge
ATOM_TYPES = {
    1: AtomType('H', 'white', 0.2, 1),
    2: AtomType('He', 'cyan', 0.286, 0),
    3: AtomType('Li', 'lightyellow', 0.34, 1),
    4: AtomType('Be', 'lightgreen', 0.589, 2),
    5: AtomType('B', 'brown', 0.415, 3),
    6: AtomType('C', 'dimgrey', 0.4, 4),
    7: AtomType('N', 'blue', 0.4, 3),
    8: AtomType('O', 'red', 0.4, 2),
    9: AtomType('F', 'purple', 0.32, 1),
    10: AtomType('Ne', 'cyan', 0.423, 0),
    11: AtomType('Na', 'lightblue', 0.485, 1),
    12: AtomType('Mg', 'darkgreen', 0.55, 2),
    13: AtomType('Al', 'lightgrey', 0.675, 3),
    14: AtomType('Si', 'darkgrey', 0.6, 4),
    15: AtomType('P', 'orange', 0.525, 3),
    16: AtomType('S', 'yellow', 0.51, 2),
    17: AtomType('Cl', 'green', 0.495, 1),
    18: AtomType('Ar', 'cyan', 0.508, 0),
    19: AtomType('K', 'navy', 0.665, 1),
    20: AtomType('Ca', 'snow', 0.495, 2),
    21: AtomType('Sc', 'lightblue', 0.72, 3),
    26: AtomType('Fe', 'darkorange', 0.67, 6),
    35: AtomType('Br', 'orange', 0.605, 1),
    36: AtomType('Kr', 'cyan', 0.524, 0),
    53: AtomType('I', 'violet', 0.7, 1),
}

ATOM_X = AtomType('X', 'black', 1, 0)


class Atom:
    def __init__(
        self,
        atomic_number: int,
        center: NDArray[np.floating],
    ) -> None:
        self.atom_type = ATOM_TYPES.get(atomic_number, ATOM_X)
        if self.atom_type is ATOM_X:
            logger.warning(
                "Invalid atomic number: %d. Atom type could not be determined. Using atom 'X' instead.",
                atomic_number,
            )

        self.center = np.array(center)
        self.mesh = pv.Sphere(center=center, radius=self.atom_type.radius)
        self.bonds: list[Bond] = []

    def remove_extra_bonds(self) -> None:
        """Remove the longest bonds if there are more bonds than `max_bonds`."""
        if len(self.bonds) <= self.atom_type.max_bonds:
            return

        self.bonds.sort(key=lambda x: x.length)

        for bond in self.bonds[self.atom_type.max_bonds :]:
            bond.mesh = None


class Bond:
    def __init__(self, atom_a: Atom, atom_b: Atom, radius: float = 0.15) -> None:
        center = (atom_a.center + atom_b.center) / 2

        bond_vec = atom_a.center - atom_b.center
        length = cast(float, np.linalg.norm(bond_vec))

        self.length = length

        self.mesh = pv.Cylinder(radius=radius, center=center, height=length, direction=bond_vec)
        self.atom_a = atom_a
        self.atom_b = atom_b

        self.color = 'grey'
        self.plotted = False

    def trim_ends(self) -> None:
        """Remove the ends of the bond that are going into the atoms."""
        if self.mesh is None:
            return

        self.mesh = self.mesh.triangulate() - self.atom_a.mesh - self.atom_b.mesh

        if self.mesh.n_points == 0:
            logger.warning(
                'Error: Bond mesh is empty between atoms %s and %s.',
                self.atom_a.atom_type.name,
                self.atom_b.atom_type.name,
            )
            self.mesh = None


class Molecule:
    def __init__(self, atoms: list[_Atom], max_bond_length: float = 4) -> None:
        # Max_bond_length helps the program skip over any bonds that should not exist
        self.max_bond_length = max_bond_length

        # Max radius is used later for plotting
        self.max_radius = 0

        self.get_atoms(atoms)

    def get_atoms(self, atoms: list[_Atom]) -> None:
        atomic_numbers = [atom.atomic_number for atom in atoms]
        atom_centers = [atom.position for atom in atoms]
        self.atoms = list(map(Atom, atomic_numbers, atom_centers))
        self.max_radius = np.max(np.linalg.norm(atom_centers, axis=1))

        distances = squareform(pdist(atom_centers))  # Compute pairwise distances
        mask = np.triu(np.ones_like(distances, dtype=bool), k=1)  # Ensure boolean mask
        indices = np.where((distances < self.max_bond_length) & mask)  # Apply mask

        for atom_a_ind, atom_b_ind in zip(indices[0], indices[1]):
            bond = Bond(self.atoms[atom_a_ind], self.atoms[atom_b_ind])
            self.atoms[atom_a_ind].bonds.append(bond)
            self.atoms[atom_b_ind].bonds.append(bond)

        for atom in self.atoms:
            atom.remove_extra_bonds()

    def add_meshes(self, plotter: pv.Plotter, opacity: float = 1) -> list[pv.Actor]:
        actors = []
        for atom in self.atoms:
            actors.append(plotter.add_mesh(atom.mesh, color=atom.atom_type.color))
            for bond in atom.bonds:
                if bond.plotted or bond.mesh is None:
                    continue

                bond.trim_ends()
                actors.append(plotter.add_mesh(bond.mesh, color=bond.color, opacity=opacity))
                bond.plotted = True

        return actors



"""Read and parse a molden file."""

import logging
from dataclasses import dataclass
from pathlib import Path

import numpy as np
from numpy.typing import NDArray
from scipy.special import gamma

logger = logging.getLogger(__name__)


@dataclass
class _Atom:
    label: str
    atomic_number: int
    position: NDArray[np.floating]
    shells: list['_Shell']


@dataclass
class _MolecularOrbital:
    sym: str
    energy: float
    spin: str
    occ: int
    coeffs: NDArray[np.floating]


class _GTO:
    def __init__(self, exp: float, coeff: float) -> None:
        self.exp = exp
        self.coeff = coeff

        self.norm = 0.0

    def normalize(self, l: int) -> None:
        # See (Jiyun Kuang and C D Lin 1997 J. Phys. B: At. Mol. Opt. Phys. 30 2529)
        # page 2532 for the normalization factor
        self.norm = np.sqrt(2 * (2 * self.exp) ** (l + 1.5) / gamma(l + 1.5))


class _Shell:
    def __init__(self, l: int, gtos: list[_GTO]) -> None:
        self.l = l
        self.gtos = gtos

        self.norm = 0.0

    def normalize(self) -> None:
        # See (Jiyun Kuang and C D Lin 1997 J. Phys. B: At. Mol. Opt. Phys. 30 2529)
        # equation 18 and 20 for the normalization factor
        for gto in self.gtos:
            gto.normalize(self.l)

        overlap = 0.0
        for i_gto in self.gtos:
            for j_gto in self.gtos:
                overlap += (
                    i_gto.coeff
                    * j_gto.coeff
                    * (2 * np.sqrt(i_gto.exp * j_gto.exp) / (i_gto.exp + j_gto.exp)) ** (self.l + 1.5)
                )

        self.norm = 1 / np.sqrt(overlap)


class Parser:
    """Parser for molden files.

    Args
    ----
        source: str | list[str]
            The path to the molden file, or the lines from the file.

        only_molecule: bool, optional
            Only parse the atoms and skip molecular orbitals.
            Default is `False`.
    """

    ANGSTROM_TO_BOHR = 1.8897259886

    def __init__(
        self,
        source: str | list[str],
        only_molecule: bool = False,
    ) -> None:
        """Initialize the Parser with either a filename or molden lines."""
        if isinstance(source, str):
            with Path(source).open('r') as file:
                self.molden_lines = file.readlines()
        elif isinstance(source, list):
            self.molden_lines = source

        # Remove leading/trailing whitespace and newline characters
        self.molden_lines = [line.strip() for line in self.molden_lines]

        self.check_molden_format()

        self.atom_ind, self.gto_ind, self.mo_ind = self.divide_molden_lines()

        self.atoms = self.get_atoms()

        if only_molecule:
            return

        self.shells = self.get_shells()
        self.mos = self.get_mos()

    def check_molden_format(self) -> None:
        """Check if the provided molden lines conform to the expected format.

        Raises
        ------
            ValueError: If the molden lines do not contain the required sections
                        or if they are in an unsupported format.

        """
        logger.info('Checking molden format...')
        if not self.molden_lines:
            raise ValueError('The provided molden lines are empty.')

        if not any('[Atoms]' in line for line in self.molden_lines):
            raise ValueError("No '[Atoms]' section found in the molden file.")

        if not any('[GTO]' in line for line in self.molden_lines):
            raise ValueError("No '[GTO]' section found in the molden file.")

        if not any('[MO]' in line for line in self.molden_lines):
            raise ValueError("No '[MO]' section found in the molden file.")

        if not any(orbs in line for orbs in ['5D', '9G'] for line in self.molden_lines):
            raise ValueError('Cartesian orbitals functions are not currently supported.')

        logger.info('Molden format check passed.')

    def divide_molden_lines(self) -> tuple[int, int, int]:
        """Divide the molden lines into sections for atoms, GTOs, and MOs.

        Returns
        -------
            tuple[int, int, int]: Indices of the '[Atoms]', '[GTO]', and '[MO]' lines.

        Raises
        ------
            ValueError: If the molden lines do not contain the required sections.

        """
        logger.info('Dividing molden lines into sections...')
        if '[Atoms] AU' in self.molden_lines:
            atom_ind = self.molden_lines.index('[Atoms] AU')
        elif '[Atoms] Angs' in self.molden_lines:
            atom_ind = self.molden_lines.index('[Atoms] Angs')
        else:
            raise ValueError("No '[Atoms] (AU/Angs)' section found in the molden file.")

        gto_ind = self.molden_lines.index('[GTO]')

        mo_ind = self.molden_lines.index('[MO]')

        logger.info('Finished dividing molden lines.')
        return atom_ind, gto_ind, mo_ind

    def get_atoms(self) -> list[_Atom]:
        """Parse the atoms from the molden file.

        Returns
        -------
            list[Atom]: A list of Atom objects containing the label, atomic number,
            and position for each atom.

        """
        logger.info('Parsing atoms...')
        angs = 'Angs' in self.molden_lines[self.atom_ind]

        atoms = []
        for line in self.molden_lines[self.atom_ind + 1 : self.gto_ind]:
            label, _, atomic_number, *coords = line.split()

            position = np.array([float(coord) for coord in coords], dtype=float)
            if angs:
                position *= self.ANGSTROM_TO_BOHR

            atoms.append(_Atom(label, int(atomic_number), position, []))

        logger.info('Parsed %s atoms.', len(atoms))
        return atoms

    def get_shells(self) -> list[_Shell]:
        """Parse the Gaussian-type orbitals (GTOs) from the molden file.

        Returns
        -------
            list[_Shell]: A list of `_Shell` objects containing the atom, angular
            momentum quantum number (l), and GTOs for each shell.

        Raises
        ------
            ValueError: If the shell label is not supported or if the GTOs are not
            formatted correctly in the molden file.

        """
        logger.info('Parsing GTO lines...')

        shell_lables = ['s', 'p', 'd', 'f', 'g']

        lines = iter(self.molden_lines[self.gto_ind + 1 : self.mo_ind])

        shells = []
        for atom in self.atoms:
            logger.debug('Parsing GTOs for atom: %s', atom.label)
            _ = next(lines)  # Skip atom index

            # Read shells until a blank line
            while True:
                line = next(lines)
                if not line:
                    break

                shell_label, num_gtos, _ = line.split()
                if shell_label not in shell_lables:
                    raise ValueError(f"Shell label '{shell_label}' is currently not supported.")

                gtos = []
                for _ in range(int(num_gtos)):
                    exp, coeff = next(lines).split()
                    gtos.append(_GTO(float(exp), float(coeff)))

                shell = _Shell(shell_lables.index(shell_label), gtos)
                shell.normalize()

                atom.shells.append(shell)
                shells.append(shell)

        logger.info('Parsed %s GTOs.', len(shells))
        return shells

    def get_mos(self, sort: bool = True) -> list[_MolecularOrbital]:
        """Parse the molecular orbitals (MOs) from the molden file.

        Args
        ----
            mo_inds: int, ArrayLike]
                Indices of the MOs to tabulate. If None, all MOs are tabulated.
            sort: bool
                If true (default), returns the mos sorted by energy. If false, returns the mos in the order
                given in the molden file.


        Returns
        -------
            list[MolecularOrbital]: A list of MolecularOrbital objects containing
            the symmetry, energy, and coefficients for each MO.

        """
        logger.info('Parsing MO coefficients...')

        num_total_gtos = sum(2 * gto.l + 1 for gto in self.shells)

        order = self._gto_order()

        lines = self.molden_lines[self.mo_ind + 1 :]
        total_num_mos = sum('Sym=' in line for line in lines)
        lines = iter(lines)

        mos = []
        for _ in range(total_num_mos):
            _, sym = next(lines).split()

            energy_line = next(lines)
            energy = float(energy_line.split()[1])

            _, spin = next(lines).split()

            occ_line = next(lines)
            occ = int(float(occ_line.split()[1]))

            coeffs = []
            for _ in range(num_total_gtos):
                _, coeff = next(lines).split()
                coeffs.append(coeff)

            mo = _MolecularOrbital(
                sym=sym,
                energy=energy,
                spin=spin,
                occ=occ,
                coeffs=np.array(coeffs, dtype=float)[order],
            )

            mos.append(mo)

        logger.info('Parsed MO coefficients.')

        if sort:
            mos = self.sort_mos(mos)

        return mos

    def _gto_order(self) -> list[int]:
        """Return the order of the GTOs in the molden file.

        Molden defines the order of the orbitals as m = 0, 1, -1, 2, -2, ...
        We want it to be m = -l, -l + 1, ..., l - 1, l.

        Note: For l = 1, the order is 1, -1, 0, which is different from the
        general pattern. This is handled separately.

        Returns
        -------
            list[int]: The order of the atomic orbitals.

        """
        order = []
        ind = 0
        for shell in self.shells:
            l = shell.l
            if l == 1:
                order.extend([ind + 1, ind + 2, ind])
            else:
                order.extend([ind + i for i in range(2 * l, -1, -2)])
                order.extend([ind + i for i in range(1, 2 * l, 2)])
            ind += 2 * l + 1

        return order

    @staticmethod
    def sort_mos(mos: list[_MolecularOrbital]) -> list[_MolecularOrbital]:
        """Sort a list of MOs by energy.

        Args
        ----
            mos: list[_MolecularOrbital]
                A list of `_MolecularOrbital` objects to be sorted.

        Returns
        -------
            list[_MolecularOrbital]: A new list containing the `_MolecularOrbital` objects sorted
            by their energy in ascending order.

        """
        logger.info('Sorting MOs by energy...')
        mos = sorted(mos, key=lambda mo: mo.energy)
        logger.info('MOs sorted by energy.')

        return mos
